# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from .viewer import (
    register_ipython_formatter,
)

__all__ = [
    "register_ipython_formatter",
]